#-------------------------------------------------------------------------%
#
#   R code used in the reply to Referee 1
#
#  
#-------------------------------------------------------------------------%

# NOTE: This code is not run time optimized. Problems
# have not appeared so far, but their absence cannot be guaranteed for.
#
# Content:  1) Packages required
#           2) Empirical $L^p$-quantile (p=2 corresponds to expectiles)
#           3) Estimator of Hill
#           4) Weissman quantile estimator
#           5) Proportionality constant (used in the next point)
#           6) Indirect Plug-in extreme $L^p$-quantile estimator
#           7) Direct LAWS extreme $L^p$-quantile estimator
#           8) Load data: Logreturns of the SP500 Index
#           9) Figure 1 
#          10) Optimal choice of the sample fraction k
#          11) Consistent Loss function
#          12) Calculation of realized losses
#          13) Figure 2
#          14) Figure 3.


#__________________________________
#
#      (1) Packages required:
#
#__________________________________

##Clear the working space:         
  rm(list = ls())
#
install.packages("ggplot2")
require(ggplot2)
#
library(reshape2)
#
install.packages("ggfortify")
library(ggfortify)
#
install.packages("zoo")
library(zoo)

#____________________________________________________
#
#   (2)  Empirical $L^p$-quantile $\widehat{q}_{Y,\tau}(p)$ :
#
#____________________________________________________
empirical_LpQuantile <- function(xtab, p, tau, linf, lsup)
{
       #inputs:
       #        xtab = sample of observations (x_1,..,x_n)
       #        tau = $L^p$-quantile level in (0,1)
       #          p = real number (power) >= 1
       #        [linf,lsup]: domain of optimization 
       #output: empirical $L^p$-quantile $\widehat{q}_{Y,\tau}(p)$       
# $L^p$-quantile check function :
eta <- function(y, p=2, tau=0.9) abs(tau-(y<=0))*(abs(y))^p
ex_obj_fct = function(theta, xtab, p, tau)
{  
  return(mean(eta(xtab-theta, p, tau)))
}
ex_obj_fct_tominimize <- function(theta, xtab, p, tau)
 unlist(lapply(theta, ex_obj_fct, xtab=xtab, p=p, tau=tau))
optimize(ex_obj_fct_tominimize, interval=c(linf,lsup), xtab=xtab, p=p, tau=tau)
}
#empirical_LpQuantile(xtab, p=2, tau=0.9, linf=0, lsup=10^7)

#__________________________________
#
#   (3) Hill's estimator $\widehat{\gamma}$: 
#
#__________________________________
Hill_inverse  <- function(ytab,k)
{
	  #(Extremiles: by Daouia.A, Juanary 7, 2013)
      #inputs:
      #ytab = sample of nx1 obs
      #    k  = sample fraction
      #output: res = 1/(Hill estimator) = Moment Mn1      
n=length(ytab);
ysort=sort(ytab);   
        g1=ysort[n-k];
        som1=0;
        for (i in 0:(k-1))
        {
            gi=ysort[n-i];
            som1 = som1 + (log(gi)-log(g1));
        }
        res=som1/k;
        return(res);
}

#__________________________________
#
#   (4) Weissman quantile estimator $\widehat{q}_{\tau'_n}^W(1)$: 
#
#__________________________________
Weissman_quantile <- function(ytab,pn,k)
{
       #(Extremiles: by Daouia.A, Juanary 7, 2013)
       #inputs:
       #       ytab = sample of nx1 obs
       #        pn  = order <= 1/n (\tau'_n = 1-pn & pn=1/n)
       #         k  = sample fraction varying from "1" up to "n-1" (tau_n = 1-k/n)
       #output: res = Weissman estimator of the extreme (1-pn)th quantile.
 
       n=length(ytab);
       ysort=sort(ytab);   
       gam = Hill_inverse(ytab,k);
       res=ysort[n-k]*(k/(n*pn))^(gam);
       return(res);
}

#__________________________________
#
#   (5) Proportionality constant $C^{-1}(\gamma;p)$: 
#
#__________________________________

Propor_cst <- function(gam,p)
{
       #inputs:
       #       gam  = tail-index $\gamma$
       #         p  = $L^p$-quantile level in the interval [1,2]
       #output: res = C^{-1}(gam;p) = [ \frac{gam}{B(p,gam^{-1}-p+1)} ]^gam

 res = (gam/beta(p,gam^(-1)-p+1))^gam;
 return(res);
}

#__________________________________
#
#   (6) Plug-in extreme $L^p$-quantile estimator $\widetilde{q}_{\tau'_n}^W(p)$: 
#
#__________________________________
Plug_Lp_quantile <- function(ytab,pn,k,power)
{
	   #inputs:
       #       ytab = sample of nx1 obs
       #        pn  = order <= 1/n  (\tau'_n = 1-pn & pn=1/n)
       #         k  = sample fraction varying from "1" up to "n-1" (tau_n = 1-k/n)
       #         power  = $L^p$-quantile power "p" in the interval [1,2]
       #output: res = Plug-in estimator of the extreme (1-pn)th $L^p$-quantile
       # $\widetilde{q}_{\tau'_n}^W(p) := C(\widehat{\gamma}_n;p) . \widehat{q}_{\tau'_n}^W(1)$
      	
       gam = Hill_inverse(ytab,k);
       cst = 1/Propor_cst(gam,power); 
       W = Weissman_quantile(ytab,pn,k);  
       res = cst * W;  
       return(res);
}

#__________________________________
#
#   (7) LAW extreme $L^p$-quantile estimator $\widehat{q}_{\tau'_n}^W(p)$: 
#
#__________________________________
LAW_Lp_quantile <- function(ytab,pn,k,power,linf,lsup)
{
	   #inputs:
       #       ytab = sample of nx1 obs
       #        pn  = order <= 1/n  (\tau'_n = 1-pn & pn=1/n)
       #         k  = sample fraction varying from "1" up to "n-1" (tau_n = 1-k/n)
       #         power  = $L^p$-quantile power "p" in the interval [1,2]
       #        [linf,lsup]: domain of optimization in the calculation of $\widehat{q}_{\tau_n}(p)$       
       #output: res = Least Asymmetrically Weighted $L^p$ estimator of the extreme (1-pn)th $L^p$-quantile
       # $\widehat{q}_{\tau'_n}^W(p) := (\frac{1-\tau'_n}{1-\tau_n})^{-\widehat{\gamma}_n} \widehat{q}_{\tau_n}(p)$
      	
       n=length(ytab);	
       emp_lpq = empirical_LpQuantile(ytab, p=power, tau=1-(k/n), linf=linf, lsup=lsup)$minimum
       gam = Hill_inverse(ytab,k);
       res=emp_lpq * (k/(n*pn))^(gam);
       return(res);
}

#__________________________________
#
# (8) Load data: Logreturns of the SP500 Index
#
#__________________________________
#install.packages("tseries")
library(tseries)
start="1994-11-02"; end= "2009-12-31"
aa = get.hist.quote(instrument = "^GSPC", start, end,
                quote =  "Close",
                provider = c("yahoo", "oanda"), method = NULL,
                           retclass = c("zoo", "its", "ts"), quiet = 
FALSE, drop = FALSE)
  str(aa)
#Logreturns of the SP500 Index:
ab = diff(log(aa));

#__________________________________
#
# (9) Figure 1: Logreturns of the SP500 Index
#
#__________________________________
autoplot(ab) +  geom_line(color='blue') + labs(x = "Time", y="Logreturns")

#______________________________________________
#
#  (10) Optimal choice of the sample fraction k:
#        
#______________________________________________
choice_k <- function(Vect,h=0.05)
{
       # 2 inputs:
       #       Vect = vector of n estimates
       #         h  = a factor to select so that the interval's length is approximately 2hn
       #         
       # 4 outputs: 
       #        opt_val = optimal estimate
       #        ind_opt = index of the optimal estimate 
       #        ind_min, ind_max = minimal and maximal indices of the block over which 
       #                           the standard deviation of the estimates is minimal  

n = length(Vect);

widt=2*max(1,ceiling(h*n))+1;  # longueur totale de l'intervalle : approx. 2h*n, nombre impair pour pouvoir aller rechercher la médiane
kmin=3;  # borne inférieure des k pouvant être choisis
kmax=n-widt;  # borne supérieure des k pouvant être choisis
NbSigma=kmax-kmin+1;  # nombre total de blocs à considérer


VectSd=numeric(NbSigma);
AV=numeric(NbSigma-2);
AP=numeric(NbSigma-2);

  for (i in 1:NbSigma)
  {
    VectSd[i]=sd(Vect[(i+kmin):(i+kmin+widt-1)])  # Vecteur d'écarts-type des estimations de gamma
  }
  
  VectSdC=VectSd[2:(length(VectSd)-1)]
  
  Sdmean=mean(VectSd)  
  M=as.numeric(VectSdC<Sdmean)  # critère 1 à vérifier pour le point candidat : inférieur à la moyenne
  
  for (i in 1:(NbSigma-2))
  {
    AV[i]=as.numeric(VectSdC[i]<VectSd[i])  
    AP[i]=as.numeric(VectSdC[i]<VectSd[i+2])  # critère 2 à vérifier pour le point candidat : minimum local
  }
  
  ToT=M+AV+AP  
  
  vecIndicMin=which(ToT==max(ToT)) #  les k vérifiant les deux critères
  IndicMin=vecIndicMin[1]+1  # le premier tel k, ajouter 1 car on a tronqué le vecteur VectSd
  IndicMin;
  
  IndicMed=which(Vect==median(Vect[(IndicMin+kmin):(IndicMin+kmin+widt-1)]))  # on cherche l'estimation médiane dans le bloc retenu
 ind_opt = IndicMed+kmin;  # indice de cette estimation

opt_val = Vect[IndicMed+kmin]
ind_min = IndicMin+kmin
ind_max = IndicMin+kmin+widt-1

        #Final result:      
        result <- list(opt_val=opt_val, ind_opt=ind_opt, ind_min=ind_min, ind_max=ind_max)
        return(result)    
}

#__________________________________
#
#   (11) Consistent Loss function: 
#
#__________________________________
eta <- function(p,tau, y)
{
       #output:  eta_{tau}(y;p)
       return(abs(tau-(y <= 0)) * abs(y)^p);
}

Loss <- function(p,tau, xi, r)
{
       #output:  L_{\tau}(xi,r) = eta_{tau}(r-xi;p)
       return(eta(p,tau,r-xi));
}

#__________________________________
#
# (12) Calculation of realized losses:
#
#__________________________________

 ytab = as.numeric(ab);
 #size of the global time series:
 N = length(ytab); 
 #size of the rolling windows :  
 n = 500;  
 #Number of forecast cases :
 T = N-n;
 #extreme quantile level $\tau'_n = 1-pn$: 
 pn = 1/n;
 #intermediate quantile level $\tau_n = 1-k/n$:
 kl = 2 
 ku = 100; 
 kk = kl:ku; 
 nk = length(kk);
#interval of optimization:
        linf=0; lsup=0.15;  # max(xtab) = 0.09469512
#$L^p$-quantile power p :       
p200   = 2;
                       # Point forecasts for p=2: 
                       Ltildep200    = matrix(0,T,nk);
                       Lhatp200      = matrix(0,T,nk);                                          
                       
                #Realizing observations "y_t": 
                 realobs =  0 * c(1:T);

                #Optimal forcasts (w.r.t. the sample fraction "k"):                
                 Otildep200    = 0 * c(1:T);
                 Ohatp200      = 0 * c(1:T);
                                                             
       for (t in 1:T)
               {                         
                 #Rolling window sample:
                 yytab = ytab[t:(n+t-1)]; 
                                  
                 #Realizing observation "y_t": 
                 realobs[t] = - ytab[n+t];    
                           # to be compared with the estimate of: -q_{yytab,(1/n)}(p) = q_{-yytab,1-(1/n)}(p)
                 
                 #Transformed sample:
                 xtab = - yytab;
                 #Objective : estimate q_{yytab,(1/n)}(p=1) = -  q_{-yytab,1-(1/n)}(p=1) = - q_{xtab,1-pn}(p=1)                    
               
                 # hat & tilde estimators for p=2:
                 hatp200    = 0 * c(1:nk);                        
                 tildep200  = 0 * c(1:nk);  
                                                  
                 #Calculation of the estimators of q_{xtab,1-pnn}(p):  
                 for (i in 1:nk)
                    {        
                       #sample fraction:
                       k=kk[i];                    

                       # hat & tilde estimators for p=1 and p=2:
                       hatp200[i]    = LAW_Lp_quantile(xtab,pn,k,p200,linf,lsup);                          
                       tildep200[i]  = Plug_Lp_quantile(xtab,pn,k,p200);                         
                       
                       # Loss functions for p=2: 
                       Ltildep200[t,i]    = Loss(2, 1-pn, tildep200[i], realobs[t]); 
                       Lhatp200[t,i]      = Loss(2, 1-pn, hatp200[i], realobs[t]);                        
                    } #fin de la boucle sur i=1:nk   
   
                          
                    #Determination of the optimal sample fraction (choosing k): 
                    h=0.01;                                                        
                       # hat & tilde estimators for p=2:
                       choice_hatp200    = choice_k(hatp200, h);                          
                       choice_tildep200  = choice_k(tildep200, h);  
                       
                 #optimal forecasts:                    
                 Otildep200[t] = choice_tildep200$opt_val;
                 Ohatp200[t] = choice_hatp200$opt_val;
               } #fin de la boucle sur t=1:T
                  
# Realized losses for p=2: 
                       Ltildep200b    = colMeans(Ltildep200,na.rm=TRUE);
                       Lhatp200b      = colMeans(Lhatp200,na.rm=TRUE);
                                   
#Determination of the optimal sample fraction (choosing k):                        
                       # hat & tilde estimators for p=1 and p=2:
                       choice_Lhatp200b    = choice_k(Lhatp200b, h);                          
                       choice_Ltildep200b  = choice_k(Ltildep200b, h);                                              
                       # hat & tilde estimators for p=2:
                       Lhatp200b_opt   = choice_Lhatp200b$opt_val;                          
                       Ltildep200b_opt = choice_Ltildep200b$opt_val;  
                      
# Transform the optimal forecasts with sign "-" as negative logreturns:
                 Otildep200 = -  Otildep200
                 Ohatp200 = -  Ohatp200
                 #
                 realobs = - realobs  
                 #  
                 tt=c(1:T);
       
#__________________________________
#
#      (13) Figure 2 : Evolution of the "realized loss" with "k"                    
#
#__________________________________
df <- data.frame(kk, 
Lhatp200b,
Ltildep200b
)           
names(df) <- c("Sample_fraction_k",
"Direct",
"Indirect"
)
ecm <- melt(df, id = "Sample_fraction_k")
dd <- qplot(Sample_fraction_k, value, data=ecm, geom="line", colour=variable)+geom_line(size = 2, linetype=1)
dd+scale_color_manual(values = c( 
 "Direct"='#FA8072',
 "Indirect" = '#3399ff'
 )) +
scale_x_continuous("Sample  fraction  k", limit = c(2,100))+
scale_y_continuous(name = "Realized  Loss", limit = c(6*10^(-6), 2.5*10^(-5))) 

#__________________________________
#
#     (14) Figure 3 : Optimal forecsats  from 1996-10-25  to  2009-12-31
#__________________________________
df_backtest <- data.frame(tt, realobs, 
Ohatp200, 
Otildep200)   
names(df_backtest) <- c("t", "Realizing observations",
"Direct forecasts",
"Indirect forecasts")
ecm <- melt(df_backtest, id = "t")
dd <- qplot(t, value, data=ecm, geom="line", colour=variable)+geom_line(size = 0, linetype=0)
dd+scale_color_manual(values = c("Realizing observations" = 'black', 
 "Direct forecasts"='#FA8072',
 "Indirect forecasts" = '#3399ff')) +
scale_y_continuous("Forecasts")+
scale_x_continuous("t") 



